import json
import logging
from Cryptodome.PublicKey import (RSA, ECC)
from Cryptodome.Hash import SHA256
from oidc_provider.lib.errors import (
    TokenError,
)
import jwt
from jwt.utils import (from_base64url_uint, base64url_decode, base64url_encode)
from jwt.exceptions import *
from django.conf import settings

logger = logging.getLogger(__name__)


def construct_key(jwk):
    # extract ECC key
    if jwk['kty'] == 'EC' and jwk['alg'] == 'ES256':
        logger.debug('Loading ECC key...')

        # decode public params
        x = from_base64url_uint(jwk['x'])
        y = from_base64url_uint(jwk['y'])

        # build the public key
        return ECC.construct(curve=jwk['crv'], point_x=x, point_y=y)

    # extract RSA key
    elif jwk['kty'] == 'RSA' and jwk['alg'] == 'RS256':
        logger.debug('Loading RSA key...')

        # decode public params
        modulus = from_base64url_uint(jwk['n'])
        exponent = from_base64url_uint(jwk['e'])

        # build the public key
        return RSA.construct((modulus, exponent))

    else:
        raise TokenError('algorithm not implemented!')


def verify_signature(jwk, dpop_proof, audience=None):
    try:
        pub_key = construct_key(jwk)
        jwt.decode(dpop_proof, key=pub_key.export_key(format='PEM'), algorithms=["ES256", "RS256"], audience=audience, leeway=getattr(settings, 'OIDC_DPOP_LEEWAY', 1))
    except InvalidAudienceError as e:
        logger.error(f'Audience isn\'t right: {str(e)}')
        raise TokenError('Invalid audience.')


def verify_dpop_proof(dpop_proof):
    """
    DPoP - Verifies the signed JWT after extracting the public key from the header
    :raises: TokenError if DPoP proof invalid
    :return: the decoded DPoP proof JWK
    """
    try:
        # will arrive in three segments
        headers = dpop_proof.split('.')[0]

        # get the jwk
        jwk = json.loads(base64url_decode(headers))['jwk']

        # verify DPOP proof signature
        verify_signature(jwk, dpop_proof)

    except KeyError:
        logger.error('Wrong format. JWK is not present or doesn\'t have the required parameters')
        raise TokenError('DPOP malformated.')

    except DecodeError:
        logger.error('Wrong format. DPOP proof encoding is not valid')
        raise TokenError('DPOP wrong encoding.')

    except (InvalidTokenError, InvalidSignatureError, ExpiredSignatureError, InvalidAlgorithmError,
            InvalidIssuerError, InvalidIssuedAtError, ImmatureSignatureError, InvalidKeyError) as e:
        logger.error(f'Wrong signature. {str(e)}')
        raise TokenError('Signature validation error')

    except MissingRequiredClaimError as e:
        logger.error(f'Missing claim: {str(e)}')
        raise TokenError('Missing claim.')

    except InvalidAudienceError as e:
        logger.error(f'Audience isn\'t right: {str(e)}')
        raise TokenError('Invalid audience.')

    return jwk


def verify_thumbprint(jwk, thumbprint):
    """
    verifies that the passed public key corresponds to the passed thumbprint
    :raises: TokenError if the thumbrpint does not match
    """
    other_thumbprint = compute_thumbprint(jwk)
    if thumbprint != other_thumbprint:
        raise TokenError('thumbprint_not_matching_dpop_token')


def compute_thumbprint(jwk):
    """Get a JSON Web Key and compute a thumbprint according to RFC 7638."""
    # see: https://tools.ietf.org/html/rfc7638
    try:
        # keep only required params in lexicographic order
        if jwk['alg'] == 'RS256':
            params = {key: jwk[key] for key in ['e', 'kty', 'n']}
        elif jwk['alg'] == 'ES256':
            params = {key: jwk[key] for key in ['crv', 'kty', 'x', 'y']}
        else:
            raise TokenError('algorithm not implemented!')

        # order them into a json string
        params = json.dumps(params, sort_keys=True, separators=(',', ':')).encode('utf-8')

        # hash
        h = SHA256.new()
        h.update(params)

        # encode
        thumbprint = base64url_encode(h.digest())

        return thumbprint.decode()

    except PyJWKError as e:
        logger.error(f'error calculating the thumbprint: {e}')
        raise TokenError('invalid thumbprint')



from hashlib import sha224
from random import randint
from uuid import uuid4

from django.forms import ModelForm
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from oidc_provider.models import Client, Code, Token, RSAKey


class ClientForm(ModelForm):

    class Meta:
        model = Client
        exclude = []

    def __init__(self, *args, **kwargs):
        super(ClientForm, self).__init__(*args, **kwargs)
        self.fields['client_id'].required = False
        self.fields['client_id'].widget.attrs['disabled'] = 'true'
        self.fields['client_secret'].required = False
        self.fields['client_secret'].widget.attrs['disabled'] = 'true'

    def clean_client_id(self):
        instance = getattr(self, 'instance', None)
        if instance and instance.pk:
            return instance.client_id
        else:
            return str(randint(1, 999999)).zfill(6)

    def clean_client_secret(self):
        instance = getattr(self, 'instance', None)

        secret = ''

        if instance and instance.pk:
            if (self.cleaned_data['client_type'] == 'confidential') and not instance.client_secret:
                secret = sha224(uuid4().hex.encode()).hexdigest()
            elif (self.cleaned_data['client_type'] == 'confidential') and instance.client_secret:
                secret = instance.client_secret
        else:
            if (self.cleaned_data['client_type'] == 'confidential'):
                secret = sha224(uuid4().hex.encode()).hexdigest()

        return secret


@admin.register(Client)
class ClientAdmin(admin.ModelAdmin):

    fieldsets = [
        [_(u''), {
            'fields': (
                'name', 'owner', 'client_type', 'response_types', '_redirect_uris', 'jwt_alg',
                'require_consent', 'reuse_consent'),
        }],
        [_(u'Credentials'), {
            'fields': ('client_id', 'client_secret', '_scope'),
        }],
        [_(u'Information'), {
            'fields': ('contact_email', 'website_url', 'terms_url', 'logo', 'date_created'),
        }],
        [_(u'Session Management'), {
            'fields': ('_post_logout_redirect_uris',),
        }],
    ]
    form = ClientForm
    list_display = ['name', 'client_id', 'response_type_descriptions', 'date_created']
    readonly_fields = ['date_created']
    search_fields = ['name']
    raw_id_fields = ['owner']


@admin.register(Code)
class CodeAdmin(admin.ModelAdmin):

    def has_add_permission(self, request):
        return False


@admin.register(Token)
class TokenAdmin(admin.ModelAdmin):

    def has_add_permission(self, request):
        return False


@admin.register(RSAKey)
class RSAKeyAdmin(admin.ModelAdmin):

    readonly_fields = ['kid']



import json
import logging
import re

from django.http import HttpResponse
from django.http import JsonResponse

from oidc_provider import settings
from oidc_provider.lib.errors import RegisterError
from oidc_provider.lib.utils.register import create_client
from oidc_provider.models import Token

logger = logging.getLogger(__name__)


class RegisterEndpoint(object):

    def __init__(self, request):
        self.request = request
        self.params = {}
        self._extract_params()

    def _extract_params(self):
        jsonStr = self.request.body

        paramDic = json.loads(jsonStr.decode())

        self.params['name'] = paramDic.get('client_name', None)
        if 'redirect_uris' in paramDic:
            self.params['redirect_uris'] = '\n'.join(paramDic['redirect_uris'])
        else:
            self.params['redirect_uris'] = None
        if 'response_types' in paramDic:
            self.params['response_types'] = paramDic['response_types']
        else:
            self.params['response_types'] = ['code']

        self.params['access_token'] = None
        header_dic = self.request.META
        if 'HTTP_AUTHORIZATION' in header_dic:
            auth_header = header_dic['HTTP_AUTHORIZATION']
            if re.compile('^Bearer\\s.+$').match(auth_header):
                self.params['access_token'] = auth_header.split()[1]

    def validate_params(self):
        # Make sure appropriate parameters are there
        # See: https://openid.net/specs/openid-connect-registration-1_0.html#ClientMetadata
        # We want client name (optional), and redirect URIs (required).
        # "Response type" will default to code (Authorization code flow)

        # Is this endpoint enabled in the configuration?
        if not settings.get('OIDC_REGISTRATION_ENDPOINT_ENABLED'):
            raise RegisterError('invalid_request')

        # If authorization is required, has user provided valid access token?
        if settings.get('OIDC_REGISTRATION_ENDPOINT_REQ_TOKEN'):
            if self.params['access_token'] is None:
                raise RegisterError('invalid_request')
            # Check whether token is valid
            try:
                self.token = Token.objects.get(access_token=self.params['access_token'])

                if self.token.has_expired():
                    logger.error('[Register] Token has expired: %s', self.params['access_token'])
                    raise RegisterError('invalid_token')

                if not ('openid' in self.token.scope):
                    logger.error('[Register] Missing openid scope.')
                    raise RegisterError('insufficient_scope')

            except Token.DoesNotExist:
                # logger.error('[UserInfo] Token does not exist: %s', self.params.access_token)
                raise RegisterError('invalid_token')

        # Has the user provided redirect URIs in JSON?
        if self.params['redirect_uris'] is None:
            raise RegisterError('invalid_request')

    def create_response_dic(self):
        """
        Create a dictionary with client_id, client_secret, and
        client_secret_expires_at (set to 0 at this point)
        See: https://openid.net/specs/openid-connect-registration-1_0.html#RegistrationResponse
        """
        client = create_client(redirect_uris=self.params['redirect_uris'],
                               name=self.params['name'],
                               response_types=self.params['response_types'])

        client.save()
        # At this point, no support for client secret expiration so
        # we return 0 meaning that it doesn't expire
        dic = {
            'client_id': client.client_id,
            'secret': client.client_secret,
            'client_secret': client.client_secret,
            'redirect_uris': client.redirect_uris,
            'client_secret_expires_at': '0'
        }

        return dic

    @classmethod
    def response(cls, dic):
        response = JsonResponse(dic, status=200)
        response['Cache-Control'] = 'no-store'
        response['Pragma'] = 'no-cache'

        return response

    @classmethod
    def error_response(cls, code, description, status):
        response = HttpResponse(status=status)
        error_pattern = 'error="{0}", error_description="{1}"'
        response['WWW-Authenticate'] = error_pattern.format(code, description)

        return response



import sys

import django
from . import settings as settings_default
from django.conf import settings

settings.configure(default_settings=settings_default)

django.setup()
from django.test.runner import DiscoverRunner

test_runner = DiscoverRunner(verbosity=1)

failures = test_runner.run_tests([
    'oidc_provider.tests.cases.test_authorize_endpoint',
    'oidc_provider.tests.cases.test_claims',
    'oidc_provider.tests.cases.test_commands',
    'oidc_provider.tests.cases.test_end_session_endpoint',
    'oidc_provider.tests.cases.test_introspection_endpoint',
    'oidc_provider.tests.cases.test_middleware',
    'oidc_provider.tests.cases.test_provider_info_endpoint',
    'oidc_provider.tests.cases.test_register_endpoint',
    'oidc_provider.tests.cases.test_settings',
    'oidc_provider.tests.cases.test_token_endpoint',
    'oidc_provider.tests.cases.test_userinfo_endpoint',
    'oidc_provider.tests.cases.test_utils',
])
if failures:
    sys.exit(failures)




from datetime import timedelta
import time
import uuid
import json
import jwt

from Cryptodome.PublicKey.RSA import importKey
from django.utils import dateformat, timezone
from jwkest.jwk import RSAKey as jwk_RSAKey
from jwkest.jwk import SYMKey
from jwkest.jws import JWS
from jwkest.jwt import JWT
from jwt.utils import (base64url_decode, base64url_encode)
from jwt.exceptions import PyJWKError

from oidc_provider.lib.utils.common import get_issuer, run_processing_hook
from oidc_provider.lib.utils.dpop import compute_thumbprint
from oidc_provider.lib.claims import StandardScopeClaims
from oidc_provider.models import (
    Code,
    RSAKey,
    Token,
    TokenType,
)
from oidc_provider.lib.errors import (
    TokenError,
)
from oidc_provider import settings


def create_id_token(token, user, aud, nonce='', at_hash='', request=None, scope=None):
    """
    Creates the id_token dictionary.
    See: http://openid.net/specs/openid-connect-core-1_0.html#IDToken
    Return a dic.
    """
    if scope is None:
        scope = []
    sub = settings.get('OIDC_IDTOKEN_SUB_GENERATOR', import_str=True)(user=user)

    expires_in = settings.get('OIDC_IDTOKEN_EXPIRE')

    # Convert datetimes into timestamps.
    now = int(time.time())
    iat_time = now
    exp_time = int(now + expires_in)
    user_auth_time = user.last_login or user.date_joined
    auth_time = int(dateformat.format(user_auth_time, 'U'))

    dic = {
        'iss': get_issuer(request=request),
        'sub': sub,
        'aud': str(aud),
        'exp': exp_time,
        'iat': iat_time,
        'auth_time': auth_time,
    }

    if token.token_type == TokenType.DPoP:
        dic['webid'] = dic['sub']

    if nonce:
        dic['nonce'] = str(nonce)

    if at_hash:
        dic['at_hash'] = at_hash

    # Inlude (or not) user standard claims in the id_token.
    if settings.get('OIDC_IDTOKEN_INCLUDE_CLAIMS'):
        if settings.get('OIDC_EXTRA_SCOPE_CLAIMS'):
            custom_claims = settings.get('OIDC_EXTRA_SCOPE_CLAIMS', import_str=True)(token)
            claims = custom_claims.create_response_dic()
        else:
            claims = StandardScopeClaims(token).create_response_dic()
        dic.update(claims)

    dic = run_processing_hook(
        dic, 'OIDC_IDTOKEN_PROCESSING_HOOK',
        user=user, token=token, request=request)

    return dic


def encode_id_token(payload, client):
    """
    Represent the ID Token as a JSON Web Token (JWT).
    Return a hash.
    """
    keys = get_client_alg_keys(client)
    _jws = JWS(payload, alg=client.jwt_alg)
    return _jws.sign_compact(keys)


def decode_id_token(token, client):
    """
    Represent the ID Token as a JSON Web Token (JWT).
    Return a hash.
    """
    keys = get_client_alg_keys(client)
    return JWS().verify_compact(token, keys=keys)


def client_id_from_id_token(id_token):
    """
    Extracts the client id from a JSON Web Token (JWT).
    Returns a string or None.
    """
    payload = JWT().unpack(id_token).payload()
    aud = payload.get('aud', None)
    if aud is None:
        return None
    if isinstance(aud, list):
        return aud[0]
    return aud


def create_token(user, client, scope, id_token_dic=None, token_type=TokenType.Bearer, request=None):
    """
    Create and populate a Token object.
    Return a Token object.
    """
    token = Token()
    token.user = user
    token.client = client
    token.token_type = token_type

    # build access token
    if token_type == TokenType.DPoP:
        expires_in = settings.get('OIDC_IDTOKEN_EXPIRE')
        now = int(time.time())
        iat_time = now
        exp_time = int(now + expires_in)

        # generate JWK thumbprint
        dpop_proof = request.headers.get('dpop')
        headers = jwt.get_unverified_header(dpop_proof)
        jwk_thumbprint = compute_thumbprint(headers.get('jwk'))

        # replace the access token with a DPoP version
        # https://solid.github.io/authentication-panel/solid-oidc-primer/#authorization-code-pkce-flow-step-18
        # TODO: the webid returned below is how it's done for the sub parameter in the id token.. suitable for us?
        #  - copy the function from djangoldp_account if you can
        # TODO: https://git.startinblox.com/djangoldp-packages/django-webidoidc-provider/issues/10
        token.access_token = encode_id_token({
            "webid": settings.get('OIDC_IDTOKEN_SUB_GENERATOR', import_str=True)(user=user),
            "iss": get_issuer(request=request),
            "aud": "solid",
            "cnf": {
                "jkt": jwk_thumbprint
            },
            "client_id": client.client_id,
            "jti": str(uuid.uuid4()),
            "iat": iat_time,
            "exp": exp_time
        }, client)
    else:
        token.access_token = uuid.uuid4().hex

    if id_token_dic is not None:
        token.id_token = id_token_dic

    token.refresh_token = uuid.uuid4().hex
    token.expires_at = timezone.now() + timedelta(
        seconds=settings.get('OIDC_TOKEN_EXPIRE'))
    token.scope = scope

    return token


def create_code(user, client, scope, nonce, is_authentication,
                code_challenge=None, code_challenge_method=None):
    """
    Create and populate a Code object.
    Return a Code object.
    """
    code = Code()
    code.user = user
    code.client = client

    code.code = uuid.uuid4().hex

    if code_challenge and code_challenge_method:
        code.code_challenge = code_challenge
        code.code_challenge_method = code_challenge_method

    code.expires_at = timezone.now() + timedelta(
        seconds=settings.get('OIDC_CODE_EXPIRE'))
    code.scope = scope
    code.nonce = nonce
    code.is_authentication = is_authentication

    return code


def get_client_alg_keys(client):
    """
    Takes a client and returns the set of (server) keys associated with it.
    Returns a list of private keys, although you can get the public key from them with .publickey()
    """
    if client.jwt_alg == 'RS256':
        keys = []
        for rsakey in RSAKey.objects.all():
            keys.append(jwk_RSAKey(key=importKey(rsakey.key), kid=rsakey.kid))
        if not keys:
            raise Exception('You must add at least one RSA Key.')
    elif client.jwt_alg == 'HS256':
        keys = [SYMKey(key=client.client_secret, alg=client.jwt_alg)]
    else:
        raise Exception('Unsupported key algorithm.')

    return keys



import inspect
from base64 import urlsafe_b64encode
import hashlib
import logging
from django.contrib.auth import authenticate

from django.http import JsonResponse

from oidc_provider.lib.errors import (
    TokenError,
    UserAuthError,
)
from oidc_provider.lib.utils.oauth2 import extract_client_auth
from oidc_provider.lib.utils.token import (
    create_id_token,
    create_token,
    encode_id_token,
)
from oidc_provider.lib.utils.dpop import verify_dpop_proof
from oidc_provider.models import (
    Client,
    Code,
    Token,
    TokenType
)
from oidc_provider import settings

logger = logging.getLogger(__name__)


class TokenEndpoint(object):

    def __init__(self, request):
        self.request = request
        self.params = {}
        self.user = None
        self._extract_params()

    def _extract_params(self):
        client_id, client_secret = extract_client_auth(self.request)

        self.params['client_id'] = client_id
        self.params['client_secret'] = client_secret
        self.params['redirect_uri'] = self.request.POST.get('redirect_uri', '')
        self.params['grant_type'] = self.request.POST.get('grant_type', '')
        self.params['code'] = self.request.POST.get('code', '')
        self.params['state'] = self.request.POST.get('state', '')
        self.params['scope'] = self.request.POST.get('scope', '')
        self.params['refresh_token'] = self.request.POST.get('refresh_token', '')
        # PKCE parameter.
        self.params['code_verifier'] = self.request.POST.get('code_verifier')
        self.params['dpop'] = self.request.headers.get('dpop')

        self.params['username'] = self.request.POST.get('username', '')
        self.params['password'] = self.request.POST.get('password', '')

    def validate_params(self):
        try:
            self.client = Client.objects.get(client_id=self.params['client_id'])
        except Client.DoesNotExist:
            logger.debug('[Token] Client does not exist: %s', self.params['client_id'])
            raise TokenError('invalid_client')

        if self.client.client_type == 'confidential':
            if not (self.client.client_secret == self.params['client_secret']):
                logger.debug('[Token] Invalid client secret: client %s do not have secret %s',
                             self.client.client_id, self.client.client_secret)
                raise TokenError('invalid_client')

        if self.params['grant_type'] == 'authorization_code':
            if not (self.params['redirect_uri'] in self.client.redirect_uris):
                logger.debug('[Token] Invalid redirect uri: %s', self.params['redirect_uri'])
                raise TokenError('invalid_client')

            try:
                self.code = Code.objects.get(code=self.params['code'])
            except Code.DoesNotExist:
                logger.debug('[Token] Code does not exist: %s', self.params['code'])
                raise TokenError('invalid_grant')

            if not (self.code.client == self.client) \
               or self.code.has_expired():
                logger.debug('[Token] Invalid code: invalid client or code has expired')
                raise TokenError('invalid_grant')

            # Validate PKCE parameters.
            if self.params['code_verifier']:
                if self.code.code_challenge_method == 'S256':
                    new_code_challenge = urlsafe_b64encode(
                            hashlib.sha256(self.params['code_verifier'].encode('ascii')).digest()
                        ).decode('utf-8').replace('=', '')
                else:
                    new_code_challenge = self.params['code_verifier']

                if not (new_code_challenge == self.code.code_challenge):
                    raise TokenError('Unable to validate PKCE parameters! Code challenge failed')

            # Validate DPoP token.
            if self.params['dpop']:
                # validate signature
                verify_dpop_proof(self.request.headers.get('dpop'))

        elif self.params['grant_type'] == 'password':
            if not settings.get('OIDC_GRANT_TYPE_PASSWORD_ENABLE'):
                raise TokenError('unsupported_grant_type')

            auth_args = (self.request,)
            try:
                inspect.getcallargs(authenticate, *auth_args)
            except TypeError:
                auth_args = ()

            user = authenticate(
                *auth_args,
                username=self.params['username'],
                password=self.params['password']
            )

            if not user:
                raise UserAuthError()

            self.user = user

        elif self.params['grant_type'] == 'refresh_token':
            if not self.params['refresh_token']:
                logger.debug('[Token] Missing refresh token')
                raise TokenError('invalid_grant')

            try:
                self.token = Token.objects.get(refresh_token=self.params['refresh_token'],
                                               client=self.client)

            except Token.DoesNotExist:
                logger.debug(
                    '[Token] Refresh token does not exist: %s', self.params['refresh_token'])
                raise TokenError('invalid_grant')
        elif self.params['grant_type'] == 'client_credentials':
            if not self.client._scope:
                logger.debug('[Token] Client using client credentials with empty scope')
                raise TokenError('invalid_scope')
        else:
            logger.debug('[Token] Invalid grant type: %s', self.params['grant_type'])
            raise TokenError('unsupported_grant_type')

    def create_response_dic(self):
        if self.params['grant_type'] == 'authorization_code':
            return self.create_code_response_dic()
        elif self.params['grant_type'] == 'refresh_token':
            return self.create_refresh_response_dic()
        elif self.params['grant_type'] == 'password':
            return self.create_access_token_response_dic()
        elif self.params['grant_type'] == 'client_credentials':
            return self.create_client_credentials_response_dic()

    def _build_response_dictionary(self, token, id_token_dic=None, scope=None):
        return_dic = {
            'access_token': token.access_token,
            'refresh_token': token.refresh_token,
            'token_type': token.token_type,
            'expires_in': settings.get('OIDC_TOKEN_EXPIRE'),
        }

        if id_token_dic is not None:
            return_dic['id_token'] = encode_id_token(id_token_dic, token.client)

        if scope is not None:
            return_dic['scope'] = scope

        return return_dic

    def _get_token_type(self):
        if self.params['dpop']:
            return TokenType.DPoP
        return TokenType.Bearer

    def _get_default_aud(self):
        # TODO: https://git.startinblox.com/djangoldp-packages/django-webidoidc-provider/issues/10
        '''if self.params['dpop']:
            return self.client.website_url'''
        return self.client.client_id

    def create_code_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-4.1

        token = create_token(
            user=self.code.user,
            client=self.code.client,
            scope=self.code.scope,
            token_type=self._get_token_type(),
            request=self.request
        )

        if self.code.is_authentication:
            id_token_dic = create_id_token(
                user=self.code.user,
                aud=self._get_default_aud(),
                token=token,
                nonce=self.code.nonce,
                at_hash=token.at_hash,
                request=self.request,
                scope=token.scope,
            )
        else:
            id_token_dic = {}
        token.id_token = id_token_dic

        # Store the token.
        token.save()

        # We don't need to store the code anymore.
        self.code.delete()

        return self._build_response_dictionary(token, id_token_dic)

    def create_refresh_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-6

        scope_param = self.params['scope']
        scope = (scope_param.split(' ') if scope_param else self.token.scope)
        unauthorized_scopes = set(scope) - set(self.token.scope)
        if unauthorized_scopes:
            raise TokenError('invalid_scope')

        token = create_token(
            user=self.token.user,
            client=self.token.client,
            scope=scope,
            token_type=self._get_token_type(),
            request=self.request
        )

        # If the Token has an id_token it's an Authentication request.
        if self.token.id_token:
            id_token_dic = create_id_token(
                user=self.token.user,
                aud=self._get_default_aud(),
                token=token,
                nonce=None,
                at_hash=token.at_hash,
                request=self.request,
                scope=token.scope,
            )
        else:
            id_token_dic = {}
        token.id_token = id_token_dic

        # Store the token.
        token.save()

        # Forget the old token.
        self.token.delete()

        return self._build_response_dictionary(token, id_token_dic)

    def create_access_token_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-4.3

        token = create_token(
            self.user,
            self.client,
            self.params['scope'].split(' '),
            token_type=self._get_token_type(),
            request=self.request
        )

        id_token_dic = create_id_token(
            token=token,
            user=self.user,
            aud=self._get_default_aud(),
            nonce='self.code.nonce',
            at_hash=token.at_hash,
            request=self.request,
            scope=token.scope,
        )

        token.id_token = id_token_dic
        token.save()

        return self._build_response_dictionary(token, id_token_dic)

    def create_client_credentials_response_dic(self):
        # See https://tools.ietf.org/html/rfc6749#section-4.4.3

        token = create_token(
            user=None,
            client=self.client,
            scope=self.client.scope,
            token_type=self._get_token_type(),
            request=self.request
        )

        token.save()

        return_dic = self._build_response_dictionary(token, scope=self.client._scope)
        return_dic.pop('refresh_token')
        return return_dic

    @classmethod
    def response(cls, dic, status=200):
        """
        Create and return a response object.
        """
        response = JsonResponse(dic, status=status)
        response['Cache-Control'] = 'no-store'
        response['Pragma'] = 'no-cache'

        return response



import uuid

from oidc_provider.models import Client, ResponseType


def create_client(redirect_uris=None, name=None, response_types=['code']):
    client = Client()
    client._redirect_uris = redirect_uris
    client.name = name or uuid.uuid4().hex
    client.client_id = uuid.uuid4().hex
    client.client_secret = uuid.uuid4().hex
    client.save()

    client.response_types.set(ResponseType.objects.filter(value__in=response_types).all())

    return client



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-01-11 18:44
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0006_unique_user_client'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='client',
            options={'verbose_name': 'Client', 'verbose_name_plural': 'Clients'},
        ),
        migrations.AlterModelOptions(
            name='code',
            options={'verbose_name': 'Authorization Code', 'verbose_name_plural': 'Authorization Codes'},
        ),
        migrations.AlterModelOptions(
            name='token',
            options={'verbose_name': 'Token', 'verbose_name_plural': 'Tokens'},
        ),
        migrations.AddField(
            model_name='client',
            name='date_created',
            field=models.DateField(
                auto_now_add=True, default=datetime.datetime(2016, 1, 11, 18, 44, 32, 192477, tzinfo=utc)),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='client',
            name='_redirect_uris',
            field=models.TextField(default=b'', help_text='Enter each URI on a new line.', verbose_name='Redirect URI'),
        ),
    ]



# Generated by Django 2.2.18 on 2021-02-15 16:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0026_client_multiple_response_types'),
    ]

    operations = [
        migrations.AddField(
            model_name='token',
            name='token_type',
            field=models.CharField(choices=[('DPoP', 'Demonstrating Proof of Posession at the Application Layer Token'), ('bearer', 'Bearer Token')], default='bearer', max_length=50),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-05 15:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0018_hybridflow_and_clientattrs'),
    ]

    operations = [
        migrations.AlterField(
            model_name='client',
            name='client_secret',
            field=models.CharField(blank=True, max_length=255, verbose_name='Client SECRET'),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0027_token_token_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='access_token',
            field=models.TextField(unique=True, verbose_name='Access Token'),
        ),
    ]



import importlib
import random
import string

from django.conf import settings


class DefaultSettings(object):
    required_attrs = ()

    def __init__(self):
        self._unauthenticated_session_management_key = None

    @property
    def OIDC_LOGIN_URL(self):
        """
        REQUIRED. Used to log the user in. By default Django's LOGIN_URL will be used.
        """
        return settings.LOGIN_URL

    @property
    def SITE_URL(self):
        """
        OPTIONAL. The OP server url.
        """
        return None

    @property
    def OIDC_AFTER_USERLOGIN_HOOK(self):
        """
        OPTIONAL.  Provide a way to plug into the process after
        the user has logged in, typically to perform some business logic.
        """
        return 'oidc_provider.lib.utils.common.default_after_userlogin_hook'

    @property
    def OIDC_AFTER_END_SESSION_HOOK(self):
        """
        OPTIONAL.  Provide a way to plug into the end session process just before calling
         Django's logout function, typically to perform some business logic.
        """
        return 'oidc_provider.lib.utils.common.default_after_end_session_hook'

    @property
    def OIDC_CODE_EXPIRE(self):
        """
        OPTIONAL. Code expiration time expressed in seconds.
        """
        return 60*10

    @property
    def OIDC_EXTRA_SCOPE_CLAIMS(self):
        """
        OPTIONAL. A string with the location of your class.
        Used to add extra scopes specific for your app.
        """
        return None

    @property
    def OIDC_IDTOKEN_EXPIRE(self):
        """
        OPTIONAL. Id token expiration time expressed in seconds.
        """
        return 60*10

    @property
    def OIDC_IDTOKEN_SUB_GENERATOR(self):
        """
        OPTIONAL. Subject Identifier. A locally unique and never
        reassigned identifier within the Issuer for the End-User,
        which is intended to be consumed by the Client.
        """
        return 'oidc_provider.lib.utils.common.default_sub_generator'

    @property
    def OIDC_IDTOKEN_INCLUDE_CLAIMS(self):
        """
        OPTIONAL. If enabled, id_token will include standard claims of the user.
        """
        return False

    @property
    def OIDC_SESSION_MANAGEMENT_ENABLE(self):
        """
        OPTIONAL. If enabled, the Server will support Session Management 1.0 specification.
        """
        return False

    @property
    def OIDC_UNAUTHENTICATED_SESSION_MANAGEMENT_KEY(self):
        """
        OPTIONAL. Supply a fixed string to use as browser-state key for unauthenticated clients.
        """

        # Memoize generated value
        if not self._unauthenticated_session_management_key:
            self._unauthenticated_session_management_key = ''.join(
                random.choice(string.ascii_uppercase + string.digits) for _ in range(100))
        return self._unauthenticated_session_management_key

    @property
    def OIDC_SKIP_CONSENT_EXPIRE(self):
        """
        OPTIONAL. User consent expiration after been granted.
        """
        return 30*3

    @property
    def OIDC_TOKEN_EXPIRE(self):
        """
        OPTIONAL. Token object expiration after been created.
        Expressed in seconds.
        """
        return 60*60

    @property
    def OIDC_USERINFO(self):
        """
        OPTIONAL. A string with the location of your function.
        Used to populate standard claims with your user information.
        """
        return 'oidc_provider.lib.utils.common.default_userinfo'

    @property
    def OIDC_IDTOKEN_PROCESSING_HOOK(self):
        """
        OPTIONAL. A string with the location of your hook.
        Used to add extra dictionary values specific for your app into id_token.
        """
        return 'oidc_provider.lib.utils.common.default_idtoken_processing_hook'

    @property
    def OIDC_INTROSPECTION_PROCESSING_HOOK(self):
        """
        OPTIONAL. A string with the location of your function.
        Used to update the response for a valid introspection token request.
        """
        return 'oidc_provider.lib.utils.common.default_introspection_processing_hook'

    @property
    def OIDC_INTROSPECTION_VALIDATE_AUDIENCE_SCOPE(self):
        """
        OPTIONAL: A boolean to specify whether or not to verify that the introspection
        resource has the requesting client id as one of its scopes.
        """
        return True

    @property
    def OIDC_GRANT_TYPE_PASSWORD_ENABLE(self):
        """
        OPTIONAL. A boolean to set whether to allow the Resource Owner Password
        Credentials Grant. https://tools.ietf.org/html/rfc6749#section-4.3

        From the specification:
            Since this access token request utilizes the resource owner's
            password, the authorization server MUST protect the endpoint
            against brute force attacks (e.g., using rate-limitation or
            generating alerts).

        How you do this, is up to you.
        """
        return False

    @property
    def OIDC_TEMPLATES(self):
        return {
            'authorize': 'oidc_provider/authorize.html',
            'error': 'oidc_provider/error.html'
        }

    @property
    def OIDC_REGISTRATION_ENDPOINT_ENABLED(self):
        """
        OPTIONAL. True if dynamic client registration endpoint is enabled
        https://openid.net/specs/openid-connect-registration-1_0.html#ClientRegistration
        """
        return True

    @property
    def OIDC_REGISTRATION_ENDPOINT_REQ_TOKEN(self):
        """
        OPTIONAL. True if client registration requires bearer (access) token.
        False if clients can be dynamically registered without authentication
        http://tools.ietf.org/html/rfc6750#section-2.1
        """
        return True

    @property
    def OIDC_REGISTRATION_ENDPOINT_ALLOW_HTTP_ORIGIN(self):
        """
        OPTIONAL. True if client registration required CORS headers.
        Useful if the Relying Party is a browser applications
        """
        return False

    @property
    def OIDC_ACCESS_CONTROL_ALLOW_HEADERS(self):
        """
        OPTIONAL.
        """
        return "Content-Type, if-match"


default_settings = DefaultSettings()


def import_from_str(value):
    """
    Attempt to import a class from a string representation.
    """
    try:
        parts = value.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError as e:
        msg = 'Could not import %s for settings. %s: %s.' % (value, e.__class__.__name__, e)
        raise ImportError(msg)


def get(name, import_str=False):
    """
    Helper function to use inside the package.
    """
    value = None
    default_value = getattr(default_settings, name)

    try:
        value = getattr(settings, name)
    except AttributeError:
        if name in default_settings.required_attrs:
            raise Exception('You must set ' + name + ' in your settings.')

    if isinstance(default_value, dict) and value:
        default_value.update(value)
        value = default_value
    else:
        if value is None:
            value = default_value
        value = import_from_str(value) if import_str else value

    return value



# -*- coding: utf-8 -*-
import base64
import binascii
from hashlib import md5, sha256
import json

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.conf import settings


CLIENT_TYPE_CHOICES = [
    ('confidential', 'Confidential'),
    ('public', 'Public'),
]

RESPONSE_TYPE_CHOICES = [
    ('code', 'code (Authorization Code Flow)'),
    ('id_token', 'id_token (Implicit Flow)'),
    ('id_token token', 'id_token token (Implicit Flow)'),
    ('code token', 'code token (Hybrid Flow)'),
    ('code id_token', 'code id_token (Hybrid Flow)'),
    ('code id_token token', 'code id_token token (Hybrid Flow)'),
]

JWT_ALGS = [
    ('HS256', 'HS256'),
    ('RS256', 'RS256'),
]


class ResponseTypeManager(models.Manager):
    def get_by_natural_key(self, value):
        return self.get(value=value)


class ResponseType(models.Model):
    objects = ResponseTypeManager()

    value = models.CharField(
        max_length=30,
        choices=RESPONSE_TYPE_CHOICES,
        unique=True,
        verbose_name=_(u'Response Type Value'))
    description = models.CharField(
        max_length=50,
    )

    def natural_key(self):
        return self.value,  # natural_key must return tuple

    def __str__(self):
        return u'{0}'.format(self.description)


class Client(models.Model):

    name = models.CharField(max_length=100, default='', verbose_name=_(u'Name'))
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_(u'Owner'), blank=True,
        null=True, default=None, on_delete=models.SET_NULL, related_name='oidc_clients_set')
    client_type = models.CharField(
        max_length=30,
        choices=CLIENT_TYPE_CHOICES,
        default='confidential',
        verbose_name=_(u'Client Type'),
        help_text=_(u'<b>Confidential</b> clients are capable of maintaining the confidentiality'
                    u' of their credentials. <b>Public</b> clients are incapable.'))
    client_id = models.CharField(max_length=255, unique=True, verbose_name=_(u'Client ID'))
    client_secret = models.CharField(max_length=255, blank=True, verbose_name=_(u'Client SECRET'))
    response_types = models.ManyToManyField(ResponseType)
    jwt_alg = models.CharField(
        max_length=10,
        choices=JWT_ALGS,
        default='RS256',
        verbose_name=_(u'JWT Algorithm'),
        help_text=_(u'Algorithm used to encode ID Tokens.'))
    date_created = models.DateField(auto_now_add=True, verbose_name=_(u'Date Created'))
    website_url = models.CharField(
        max_length=255, blank=True, default='', verbose_name=_(u'Website URL'))
    terms_url = models.CharField(
        max_length=255,
        blank=True,
        default='',
        verbose_name=_(u'Terms URL'),
        help_text=_(u'External reference to the privacy policy of the client.'))
    contact_email = models.CharField(
        max_length=255, blank=True, default='', verbose_name=_(u'Contact Email'))
    logo = models.FileField(
        blank=True, default='', upload_to='oidc_provider/clients', verbose_name=_(u'Logo Image'))
    reuse_consent = models.BooleanField(
        default=True,
        verbose_name=_('Reuse Consent?'),
        help_text=_('If enabled, server will save the user consent given to a specific client, '
                    'so that user won\'t be prompted for the same authorization multiple times.'))
    require_consent = models.BooleanField(
        default=True,
        verbose_name=_('Require Consent?'),
        help_text=_('If disabled, the Server will NEVER ask the user for consent.'))
    _redirect_uris = models.TextField(
        default='', verbose_name=_(u'Redirect URIs'),
        help_text=_(u'Enter each URI on a new line.'))
    _post_logout_redirect_uris = models.TextField(
        blank=True,
        default='',
        verbose_name=_(u'Post Logout Redirect URIs'),
        help_text=_(u'Enter each URI on a new line.'))
    _scope = models.TextField(
        blank=True,
        default='',
        verbose_name=_(u'Scopes'),
        help_text=_('Specifies the authorized scope values for the client app.'))

    class Meta:
        verbose_name = _(u'Client')
        verbose_name_plural = _(u'Clients')

    def __str__(self):
        return u'{0}'.format(self.name)

    def __unicode__(self):
        return self.__str__()

    def response_type_values(self):
        return (response_type.value for response_type in self.response_types.all())

    def response_type_descriptions(self):
        # return as a list, rather than a generator, so descriptions display correctly in admin
        return [response_type.description for response_type in self.response_types.all()]

    @property
    def redirect_uris(self):
        return self._redirect_uris.splitlines()

    @redirect_uris.setter
    def redirect_uris(self, value):
        self._redirect_uris = '\n'.join(value)

    @property
    def post_logout_redirect_uris(self):
        return self._post_logout_redirect_uris.splitlines()

    @post_logout_redirect_uris.setter
    def post_logout_redirect_uris(self, value):
        self._post_logout_redirect_uris = '\n'.join(value)

    @property
    def scope(self):
        return self._scope.split()

    @scope.setter
    def scope(self, value):
        self._scope = ' '.join(value)

    @property
    def default_redirect_uri(self):
        return self.redirect_uris[0] if self.redirect_uris else ''


class BaseCodeTokenModel(models.Model):

    client = models.ForeignKey(Client, verbose_name=_(u'Client'), on_delete=models.CASCADE)
    expires_at = models.DateTimeField(verbose_name=_(u'Expiration Date'))
    _scope = models.TextField(default='', verbose_name=_(u'Scopes'))

    class Meta:
        abstract = True

    @property
    def scope(self):
        return self._scope.split()

    @scope.setter
    def scope(self, value):
        self._scope = ' '.join(value)

    def __unicode__(self):
        return self.__str__()

    def has_expired(self):
        return timezone.now() >= self.expires_at


class Code(BaseCodeTokenModel):

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_(u'User'), on_delete=models.CASCADE)
    code = models.CharField(max_length=255, unique=True, verbose_name=_(u'Code'))
    nonce = models.CharField(max_length=255, blank=True, default='', verbose_name=_(u'Nonce'))
    is_authentication = models.BooleanField(default=False, verbose_name=_(u'Is Authentication?'))
    code_challenge = models.CharField(max_length=255, null=True, verbose_name=_(u'Code Challenge'))
    code_challenge_method = models.CharField(
        max_length=255, null=True, verbose_name=_(u'Code Challenge Method'))

    class Meta:
        verbose_name = _(u'Authorization Code')
        verbose_name_plural = _(u'Authorization Codes')

    def __str__(self):
        return u'{0} - {1}'.format(self.client, self.code)


class TokenType(object):
    DPoP = 'DPoP'
    Bearer = 'bearer'

    @classmethod
    def choices(cls):
        return (
            (cls.DPoP, 'Demonstrating Proof of Posession at the Application Layer Token'),
            (cls.Bearer, 'Bearer Token'),
        )


class Token(BaseCodeTokenModel):

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, verbose_name=_(u'User'), on_delete=models.CASCADE)
    access_token = models.TextField(unique=True, verbose_name=_(u'Access Token'))
    refresh_token = models.CharField(max_length=255, unique=True, verbose_name=_(u'Refresh Token'))
    _id_token = models.TextField(verbose_name=_(u'ID Token'))
    token_type = models.CharField(choices=TokenType.choices(), max_length=50, blank=False, null=False,
                                  default=TokenType.Bearer)

    class Meta:
        verbose_name = _(u'Token')
        verbose_name_plural = _(u'Tokens')

    @property
    def id_token(self):
        return json.loads(self._id_token) if self._id_token else None

    @id_token.setter
    def id_token(self, value):
        self._id_token = json.dumps(value)

    def __str__(self):
        return u'{0} - {1}'.format(self.client, self.access_token)

    @property
    def at_hash(self):
        # @@@ d-o-p only supports 256 bits (change this if that changes)
        hashed_access_token = sha256(
            self.access_token.encode('ascii')
        ).hexdigest().encode('ascii')
        return base64.urlsafe_b64encode(
            binascii.unhexlify(
                hashed_access_token[:len(hashed_access_token) // 2]
            )
        ).rstrip(b'=').decode('ascii')


class UserConsent(BaseCodeTokenModel):

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_(u'User'), on_delete=models.CASCADE)
    date_given = models.DateTimeField(verbose_name=_(u'Date Given'))

    class Meta:
        unique_together = ('user', 'client')


class RSAKey(models.Model):

    key = models.TextField(
        verbose_name=_(u'Key'), help_text=_(u'Paste your private RSA Key here.'))

    class Meta:
        verbose_name = _(u'RSA Key')
        verbose_name_plural = _(u'RSA Keys')

    def __str__(self):
        return u'{0}'.format(self.kid)

    def __unicode__(self):
        return self.__str__()

    @property
    def kid(self):
        return u'{0}'.format(md5(self.key.encode('utf-8')).hexdigest() if self.key else '')



import logging

from django.http import JsonResponse

from oidc_provider.lib.errors import TokenIntrospectionError
from oidc_provider.lib.utils.common import run_processing_hook
from oidc_provider.lib.utils.oauth2 import extract_client_auth
from oidc_provider.models import Token, Client
from oidc_provider import settings

logger = logging.getLogger(__name__)

INTROSPECTION_SCOPE = 'token_introspection'


class TokenIntrospectionEndpoint(object):

    def __init__(self, request):
        self.request = request
        self.params = {}
        self.token = None
        self.id_token = None
        self.client = None
        self._extract_params()

    def _extract_params(self):
        # Introspection only supports POST requests
        self.params['token'] = self.request.POST.get('token')
        client_id, client_secret = extract_client_auth(self.request)
        self.params['client_id'] = client_id
        self.params['client_secret'] = client_secret

    def validate_params(self):
        if not (self.params['client_id'] and self.params['client_secret']):
            logger.debug('[Introspection] No client credentials provided')
            raise TokenIntrospectionError()
        if not self.params['token']:
            logger.debug('[Introspection] No token provided')
            raise TokenIntrospectionError()
        try:
            self.token = Token.objects.get(access_token=self.params['token'])
        except Token.DoesNotExist:
            logger.debug('[Introspection] Token does not exist: %s', self.params['token'])
            raise TokenIntrospectionError()
        if self.token.has_expired():
            logger.debug('[Introspection] Token is not valid: %s', self.params['token'])
            raise TokenIntrospectionError()

        try:
            self.client = Client.objects.get(
                client_id=self.params['client_id'],
                client_secret=self.params['client_secret'])
        except Client.DoesNotExist:
            logger.debug('[Introspection] No valid client for id: %s',
                         self.params['client_id'])
            raise TokenIntrospectionError()
        if INTROSPECTION_SCOPE not in self.client.scope:
            logger.debug('[Introspection] Client %s does not have introspection scope',
                         self.params['client_id'])
            raise TokenIntrospectionError()

        self.id_token = self.token.id_token

        if settings.get('OIDC_INTROSPECTION_VALIDATE_AUDIENCE_SCOPE'):
            if not self.token.id_token:
                logger.debug('[Introspection] Token not an authentication token: %s',
                             self.params['token'])
                raise TokenIntrospectionError()

            audience = self.token.id_token.get('aud')
            if not audience:
                logger.debug('[Introspection] No audience found for token: %s',
                             self.params['token'])
                raise TokenIntrospectionError()

            if audience not in self.client.scope:
                logger.debug('[Introspection] Client %s does not audience scope %s',
                             self.params['client_id'], audience)
                raise TokenIntrospectionError()

    def create_response_dic(self):
        response_dic = {}
        if self.id_token:
            for k in ('aud', 'sub', 'exp', 'iat', 'iss'):
                response_dic[k] = self.id_token[k]
        response_dic['active'] = True
        response_dic['client_id'] = self.token.client.client_id

        response_dic = run_processing_hook(response_dic,
                                           'OIDC_INTROSPECTION_PROCESSING_HOOK',
                                           client=self.client,
                                           id_token=self.id_token)

        return response_dic

    @classmethod
    def response(cls, dic, status=200):
        """
        Create and return a response object.
        """
        response = JsonResponse(dic, status=status)
        response['Cache-Control'] = 'no-store'
        response['Pragma'] = 'no-cache'

        return response



try:
    # https://docs.djangoproject.com/en/1.10/topics/http/middleware/#upgrading-pre-django-1-10-style-middleware
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

from oidc_provider import settings
from oidc_provider.lib.utils.common import get_browser_state_or_default


class SessionManagementMiddleware(MiddlewareMixin):
    """
    Maintain a `op_browser_state` cookie along with the `sessionid` cookie that
    represents the End-User's login state at the OP. If the user is not logged
    in then use the value of settings.OIDC_UNAUTHENTICATED_SESSION_MANAGEMENT_KEY.
    """

    def process_response(self, request, response):
        if settings.get('OIDC_SESSION_MANAGEMENT_ENABLE'):
            response.set_cookie('op_browser_state', get_browser_state_or_default(request))
        return response



import random
import string
import base64
import time
import uuid
import json
import jwt
import logging
from math import floor
from jwt.utils import to_base64url_uint

import Cryptodome
from Cryptodome.PublicKey import (RSA, ECC)
from Cryptodome.Hash import SHA256
from Cryptodome.Signature import pkcs1_15


import django
from django.contrib.auth.backends import ModelBackend

try:
    from urlparse import parse_qs, urlsplit
except ImportError:
    from urllib.parse import parse_qs, urlsplit

from django.utils import timezone
from django.contrib.auth.models import User

from oidc_provider.models import (
    Client,
    Code,
    Token,
    ResponseType)

logger = logging.getLogger(__name__)

FAKE_NONCE = 'cb584e44c43ed6bd0bc2d9c7e242837d'
FAKE_RANDOM_STRING = ''.join(
    random.choice(string.ascii_uppercase + string.digits) for _ in range(32))
FAKE_CODE_CHALLENGE = 'YlYXEqXuRm-Xgi2BOUiK50JW1KsGTX6F1TDnZSC8VTg'
FAKE_CODE_VERIFIER = 'SmxGa0XueyNh5bDgTcSrqzAh2_FmXEqU8kDT6CuXicw'

# avoid computation exenstive cost
FAKE_ECP256_PRIVKEY = """-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIHdm2ynC4Cn4uqTpQ1W5kikdxrcXEk8ESAsxXPH//epLoAoGCCqGSM49
AwEHoUQDQgAEq2Wk3s1RYzO4E0lDT9+7w7IYx9NFutAMbeZj+sb+tY0HdXC2VHYb
MoJk4VfoYCuLl8orNqgFYPa/WsctXAnwYQ==
-----END EC PRIVATE KEY-----"""

FAKE_RSA4096_PRIVKEY = """-----BEGIN RSA PRIVATE KEY-----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-----END RSA PRIVATE KEY-----"""


def create_fake_user():
    """
    Create a test user.

    Return a User object.
    """
    user = User()
    user.username = 'johndoe'
    user.email = 'johndoe@example.com'
    user.first_name = 'John'
    user.last_name = 'Doe'
    user.set_password('1234')

    user.save()

    return user


def create_fake_client(response_type, is_public=False, require_consent=True):
    """
    Create a test client, response_type argument MUST be:
    'code', 'id_token' or 'id_token token'.

    Return a Client object.
    """
    client = Client()
    client.name = 'Some Client'
    client.client_id = str(random.randint(1, 999999)).zfill(6)
    if is_public:
        client.client_type = 'public'
        client.client_secret = ''
    else:
        client.client_secret = str(random.randint(1, 999999)).zfill(6)
    client.redirect_uris = ['http://example.com/']
    client.require_consent = require_consent

    client.save()

    # check if response_type is a string in a python 2 and 3 compatible way
    if isinstance(response_type, ("".__class__, u"".__class__)):
        response_type = (response_type,)
    for value in response_type:
        client.response_types.add(ResponseType.objects.get(value=value))

    return client


def create_fake_keys():
    '''
    creates a private-public key pair (oidc_provider.models.RSAKey) for use in testing
    '''
    key = RSA.generate(2048)
    return key


def sign_with_key(data, private_key):
    '''
    :param data: a bytes-like object to sign
    :return: the signed message (bytes)
    '''
    # https://pycryptodome.readthedocs.io/en/latest/src/signature/signature.html
    signer = Cryptodome.Signature.pkcs1_15.new(private_key)
    hash = Cryptodome.Hash.SHA384.new()
    hash.update(data)
    return signer.sign(hash)


def create_fake_token(user, scopes, client):
    expires_at = timezone.now() + timezone.timedelta(seconds=60)
    token = Token(user=user, client=client, expires_at=expires_at)
    token.scope = scopes

    token.save()

    return token


def create_fake_dpop_proof(private_key_header=FAKE_RSA4096_PRIVKEY, signing_key=None,
                           htu="https://secureauth.example/token", htm="POST"):
    """
    Constructs a mock DPoP Proof sent by the client.
    """

    #FIXME: PyJWT uses "cryptography" as crypto backend. Extraction and formating should use the same.
    # get the test keypair
    try:
        # build headers from RSA key
        privkey = RSA.import_key(private_key_header)

        # get public params
        modulus = privkey.n
        exponent = privkey.e

        # encode them
        modulus_b64 = to_base64url_uint(modulus).decode()
        exponent_b64 = to_base64url_uint(exponent).decode()

        # build a JWK with RSA-SHA256
        # see: https://www.rfc-editor.org/rfc/rfc7518.html#page-30
        jwk = {
          "kty": "RSA",
          "alg": "RS256",
          "use": "sig",
          "kid": 1,
          "n": modulus_b64,
          "e": exponent_b64
        }

        # may override signkey for tests
        signing_key = RSA.import_key(signing_key) if signing_key else privkey


    except ValueError:
        try:
            # build headers from ECC key
            privkey = ECC.import_key(private_key_header)

            # get public params
            # see: https://github.com/Legrandin/pycryptodome/blob/master/lib/Crypto/Math/_IntegerGMP.py#L171
            x = int(privkey.pointQ.x)
            y = int(privkey.pointQ.y)

            # encode them
            x_b64 = to_base64url_uint(x).decode()
            y_b64 = to_base64url_uint(y).decode()

            # build a JWK with EC-P256-SHA256
            # https://www.rfc-editor.org/rfc/rfc7518.html#page-28
            jwk = {
                "kty": "EC",
                "alg": "ES256",
                "use": "sig",
                "kid": 1,
                "crv": "P-256",
                "x": x_b64,
                "y": y_b64
            }

            # may override signkey for tests
            signing_key = ECC.import_key(signing_key) if signing_key else privkey

        except: raise

    # use the public to check signature with external tools
    #logger.debug(f'PUBKEY: {privkey.publickey().export_key().decode()}')

    # build a DPOP header
    headers = {
      "alg": jwk["alg"],
      "typ": "dpop+jwt",
      "jwk": jwk
    }

    # claim something
    payload = {
        "htu": htu,
        "htm": htm,
        "jti": str(uuid.uuid4()),
        "iat": int(floor(time.time()))
    }

    # build a signed JWT
    return jwt.encode(payload, signing_key.export_key(format="PEM"), algorithm=headers["alg"], headers=headers)


def is_code_valid(url, user, client):
    """
    Check if the code inside the url is valid. Supporting both query string and fragment.
    """
    try:
        parsed = urlsplit(url)
        params = parse_qs(parsed.query or parsed.fragment)
        code = params['code'][0]
        code = Code.objects.get(code=code)
        is_code_ok = (code.client == client) and (code.user == user)
    except Exception:
        is_code_ok = False

    return is_code_ok


def userinfo(claims, user):
    """
    Fake function for setting OIDC_USERINFO.
    """
    claims['given_name'] = 'John'
    claims['family_name'] = 'Doe'
    claims['name'] = '{0} {1}'.format(claims['given_name'], claims['family_name'])
    claims['email'] = user.email
    claims['email_verified'] = True
    claims['address']['country'] = 'Argentina'
    return claims


def fake_sub_generator(user):
    """
    Fake function for setting OIDC_IDTOKEN_SUB_GENERATOR.
    """
    return user.email


def fake_idtoken_processing_hook(id_token, user, **kwargs):
    """
    Fake function for inserting some keys into token. Testing OIDC_IDTOKEN_PROCESSING_HOOK.
    """
    id_token['test_idtoken_processing_hook'] = FAKE_RANDOM_STRING
    id_token['test_idtoken_processing_hook_user_email'] = user.email
    return id_token


def fake_idtoken_processing_hook2(id_token, user, **kwargs):
    """
    Fake function for inserting some keys into token.
    Testing OIDC_IDTOKEN_PROCESSING_HOOK - tuple or list as param
    """
    id_token['test_idtoken_processing_hook2'] = FAKE_RANDOM_STRING
    id_token['test_idtoken_processing_hook_user_email2'] = user.email
    return id_token


def fake_idtoken_processing_hook3(id_token, user, token, **kwargs):
    """
    Fake function for checking scope is passed to processing hook.
    """
    id_token['scope_of_token_passed_to_processing_hook'] = token.scope
    return id_token


def fake_idtoken_processing_hook4(id_token, user, **kwargs):
    """
    Fake function for checking kwargs passed to processing hook.
    """
    id_token['kwargs_passed_to_processing_hook'] = {
        key: repr(value)
        for (key, value) in kwargs.items()
    }
    return id_token


def fake_introspection_processing_hook(response_dict, client, id_token):
    response_dict['test_introspection_processing_hook'] = FAKE_RANDOM_STRING
    return response_dict


class TestAuthBackend:
    def authenticate(self, *args, **kwargs):
        if django.VERSION[0] >= 2 or (django.VERSION[0] == 1 and django.VERSION[1] >= 11):
            assert len(args) > 0 and args[0]
        return ModelBackend().authenticate(*args, **kwargs)



import os
from setuptools import (
    find_packages,
    setup,
)

version = {}
with open("./oidc_provider/version.py") as fp:
    exec(fp.read(), version)

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-webidoidc-provider',
    version=version['__version__'],
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='WEBID - OIDC Provider implementation for Django.',
    long_description='https://git.startinblox.com/djangoldp-packages/django-webidoidc-provider',
    url='https://git.startinblox.com/djangoldp-packages/django-webidoidc-provider',
    author='Juan Ignacio Fiorentino + Startin\'blox Team',
    author_email='support@startinblox.com',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    test_suite='runtests.runtests',
    tests_require=[
        'mock>=2.0.0',
    ],

    install_requires=[
        'pyjwkest>=1.3.0',
        'pyjwt[crypto]~=2.5.0',
    ]
)



from base64 import b64decode
import logging
import re

from django.http import HttpResponse

from oidc_provider.lib.errors import BearerTokenError
from oidc_provider.models import Token


logger = logging.getLogger(__name__)


def extract_access_token(request):
    """
    Get the access token using Authorization Request Header Field method.
    Or try getting via GET.
    See: http://tools.ietf.org/html/rfc6750#section-2.1

    Return a string.
    """
    auth_header = request.headers.get('authorization', '')

    if re.compile(r'^[Bb]earer\s{1}.+$').match(auth_header):
        access_token = auth_header.split()[1]
    else:
        access_token = request.GET.get('access_token', '')

    return access_token


def extract_client_auth(request):
    """
    Get client credentials using HTTP Basic Authentication method.
    Or try getting parameters via POST.
    See: http://tools.ietf.org/html/rfc6750#section-2.1

    Return a tuple `(client_id, client_secret)`.
    """
    auth_header = request.headers.get('authorization', '')

    if re.compile(r'^Basic\s{1}.+$').match(auth_header):
        b64_user_pass = auth_header.split()[1]
        try:
            user_pass = b64decode(b64_user_pass).decode('utf-8').split(':')
            client_id, client_secret = tuple(user_pass)
        except Exception:
            client_id = client_secret = ''
    else:
        client_id = request.POST.get('client_id', '')
        client_secret = request.POST.get('client_secret', '')

    return (client_id, client_secret)


def protected_resource_view(scopes=None):
    """
    View decorator. The client accesses protected resources by presenting the
    access token to the resource server.
    https://tools.ietf.org/html/rfc6749#section-7
    """
    if scopes is None:
        scopes = []

    def wrapper(view):
        def view_wrapper(request,  *args, **kwargs):
            access_token = extract_access_token(request)

            try:
                try:
                    kwargs['token'] = Token.objects.get(access_token=access_token)
                except Token.DoesNotExist:
                    logger.debug('[UserInfo] Token does not exist: %s', access_token)
                    raise BearerTokenError('invalid_token')

                if kwargs['token'].has_expired():
                    logger.debug('[UserInfo] Token has expired: %s', access_token)
                    raise BearerTokenError('invalid_token')

                if not set(scopes).issubset(set(kwargs['token'].scope)):
                    logger.debug('[UserInfo] Missing openid scope.')
                    raise BearerTokenError('insufficient_scope')
            except BearerTokenError as error:
                response = HttpResponse(status=error.status)
                response['WWW-Authenticate'] = 'error="{0}", error_description="{1}"'.format(
                    error.code, error.description)
                return response

            return view(request,  *args, **kwargs)

        return view_wrapper

    return wrapper



# Generated by Django 2.0.7 on 2018-08-15 20:44

from django.db import migrations, models


def migrate_response_type(apps, schema_editor):
    RESPONSE_TYPES = [
        ('code', 'code (Authorization Code Flow)'),
        ('id_token', 'id_token (Implicit Flow)'),
        ('id_token token', 'id_token token (Implicit Flow)'),
        ('code token', 'code token (Hybrid Flow)'),
        ('code id_token', 'code id_token (Hybrid Flow)'),
        ('code id_token token', 'code id_token token (Hybrid Flow)'),
    ]
    # ensure we get proper, versioned model with the deleted response_type field;
    # importing directly yields the latest without response_type
    ResponseType = apps.get_model('oidc_provider', 'ResponseType')
    Client = apps.get_model('oidc_provider', 'Client')
    for value, description in RESPONSE_TYPES:
        ResponseType.objects.create(value=value, description=description)
    for client in Client.objects.all():
        client.response_types.add(ResponseType.objects.get(value=client.response_type))


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0025_user_field_codetoken'),
    ]

    operations = [
        migrations.CreateModel(
            name='ResponseType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(choices=[('code', 'code (Authorization Code Flow)'), ('id_token', 'id_token (Implicit Flow)'), ('id_token token', 'id_token token (Implicit Flow)'), ('code token', 'code token (Hybrid Flow)'), ('code id_token', 'code id_token (Hybrid Flow)'), ('code id_token token', 'code id_token token (Hybrid Flow)')], max_length=30, unique=True, verbose_name='Response Type Value')),
                ('description', models.CharField(max_length=50)),
            ],
        ),
        migrations.AddField(
            model_name='client',
            name='response_types',
            field=models.ManyToManyField(to='oidc_provider.ResponseType'),
        ),
        # omitting reverse migrate_response_type since removing response_type is irreversible (nonnull and no default)
        migrations.RunPython(migrate_response_type),
        migrations.RemoveField(
            model_name='client',
            name='response_type',
        ),
    ]
